<Styles xmlns="https://github.com/avaloniaui"
        xmlns:x="http://schemas.microsoft.com/winfx/2006/xaml">
  <Styles.Resources>
    <ResourceDictionary>
      <StaticResource x:Key="SearchPanelFontSize" ResourceKey="ControlContentThemeFontSize" />
      <StaticResource x:Key="SearchPanelFontFamily" ResourceKey="ContentControlThemeFontFamily" />
      <ResourceDictionary.ThemeDictionaries>
        <ResourceDictionary x:Key="Default">
          <StaticResource x:Key="CompletionToolTipBackground" ResourceKey="ToolTipBackground" />
          <StaticResource x:Key="CompletionToolTipForeground" ResourceKey="ToolTipForeground" />
          <StaticResource x:Key="CompletionToolTipBorderBrush" ResourceKey="ToolTipBorderBrush" />
          <StaticResource x:Key="CompletionToolTipBorderThickness" ResourceKey="ToolTipBorderThemeThickness" />
          <StaticResource x:Key="OverloadViewerBackground" ResourceKey="ToolTipBackground" />
          <StaticResource x:Key="OverloadViewerForeground" ResourceKey="ToolTipForeground" />
          <StaticResource x:Key="OverloadViewerBorderBrush" ResourceKey="ToolTipBorderBrush" />
          <SolidColorBrush x:Key="SearchPanelBackgroundBrush" Color="{DynamicResource SystemChromeMediumColor}" />
          <SolidColorBrush x:Key="SearchPanelBorderBrush" Color="{DynamicResource SystemBaseLowColor}" />
          <SolidColorBrush x:Key="TextAreaSelectionBrush" Color="{DynamicResource SystemAccentColor}" />
        </ResourceDictionary>
        <ResourceDictionary x:Key="Dark">
          <StaticResource x:Key="CompletionToolTipBackground" ResourceKey="ToolTipBackground" />
          <StaticResource x:Key="CompletionToolTipForeground" ResourceKey="ToolTipForeground" />
          <StaticResource x:Key="CompletionToolTipBorderBrush" ResourceKey="ToolTipBorderBrush" />
          <StaticResource x:Key="CompletionToolTipBorderThickness" ResourceKey="ToolTipBorderThemeThickness" />
          <StaticResource x:Key="OverloadViewerBackground" ResourceKey="ToolTipBackground" />
          <StaticResource x:Key="OverloadViewerForeground" ResourceKey="ToolTipForeground" />
          <StaticResource x:Key="OverloadViewerBorderBrush" ResourceKey="ToolTipBorderBrush" />
          <SolidColorBrush x:Key="SearchPanelBackgroundBrush" Color="{DynamicResource SystemChromeMediumColor}" />
          <SolidColorBrush x:Key="SearchPanelBorderBrush" Color="{DynamicResource SystemBaseLowColor}" />
          <SolidColorBrush x:Key="TextAreaSelectionBrush" Color="{DynamicResource SystemAccentColor}" />
        </ResourceDictionary>
      </ResourceDictionary.ThemeDictionaries>

      <ResourceDictionary.MergedDictionaries>
        <ResourceInclude Source="../Base.xaml" />
      </ResourceDictionary.MergedDictionaries>
    </ResourceDictionary>
  </Styles.Resources>
</Styles>