<ResourceDictionary xmlns="https://github.com/avaloniaui"
                    xmlns:x="http://schemas.microsoft.com/winfx/2006/xaml">
  <ResourceDictionary.MergedDictionaries>
    <MergeResourceInclude Source="/TextEditor.xaml" />
    <MergeResourceInclude Source="/Editing/TextArea.xaml" />
    <MergeResourceInclude Source="/CodeCompletion/CompletionTipContentControl.xaml" />
    <MergeResourceInclude Source="/CodeCompletion/CompletionList.xaml" />
    <MergeResourceInclude Source="/CodeCompletion/InsightWindow.xaml" />
    <MergeResourceInclude Source="/Search/SearchPanel.xaml" />
  </ResourceDictionary.MergedDictionaries>
</ResourceDictionary>
