//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a tool.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------

namespace AvaloniaEdit {
    using System;
    
    
    /// <summary>
    ///   A strongly-typed resource class, for looking up localized strings, etc.
    /// </summary>
    // This class was auto-generated by the StronglyTypedResourceBuilder
    // class via a tool like ResGen or Visual Studio.
    // To add or remove a member, edit your .ResX file then rerun ResGen
    // with the /str option, or rebuild your VS project.
    [global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "16.0.0.0")]
    [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
    [global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]
    public class SR {
        
        private static global::System.Resources.ResourceManager resourceMan;
        
        private static global::System.Globalization.CultureInfo resourceCulture;
        
        [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
        internal SR() {
        }
        
        /// <summary>
        ///   Returns the cached ResourceManager instance used by this class.
        /// </summary>
        [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
        public static global::System.Resources.ResourceManager ResourceManager {
            get {
                if (object.ReferenceEquals(resourceMan, null)) {
                    global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("AvaloniaEdit.SR", typeof(SR).Assembly);
                    resourceMan = temp;
                }
                return resourceMan;
            }
        }
        
        /// <summary>
        ///   Overrides the current thread's CurrentUICulture property for all
        ///   resource lookups using this strongly typed resource class.
        /// </summary>
        [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
        public static global::System.Globalization.CultureInfo Culture {
            get {
                return resourceCulture;
            }
            set {
                resourceCulture = value;
            }
        }
        
        /// <summary>
        ///   Looks up a localized string similar to Replace....
        /// </summary>
        public static string ReplaceLabel {
            get {
                return ResourceManager.GetString("ReplaceLabel", resourceCulture);
            }
        }
        
        /// <summary>
        ///   Looks up a localized string similar to 1 match.
        /// </summary>
        public static string Search1Match {
            get {
                return ResourceManager.GetString("Search1Match", resourceCulture);
            }
        }
        
        /// <summary>
        ///   Looks up a localized string similar to Error: .
        /// </summary>
        public static string SearchErrorText {
            get {
                return ResourceManager.GetString("SearchErrorText", resourceCulture);
            }
        }
        
        /// <summary>
        ///   Looks up a localized string similar to Find next (F3).
        /// </summary>
        public static string SearchFindNextText {
            get {
                return ResourceManager.GetString("SearchFindNextText", resourceCulture);
            }
        }
        
        /// <summary>
        ///   Looks up a localized string similar to Find previous (Shift+F3).
        /// </summary>
        public static string SearchFindPreviousText {
            get {
                return ResourceManager.GetString("SearchFindPreviousText", resourceCulture);
            }
        }
        
        /// <summary>
        ///   Looks up a localized string similar to Find....
        /// </summary>
        public static string SearchLabel {
            get {
                return ResourceManager.GetString("SearchLabel", resourceCulture);
            }
        }
        
        /// <summary>
        ///   Looks up a localized string similar to Match case.
        /// </summary>
        public static string SearchMatchCaseText {
            get {
                return ResourceManager.GetString("SearchMatchCaseText", resourceCulture);
            }
        }
        
        /// <summary>
        ///   Looks up a localized string similar to Match whole words.
        /// </summary>
        public static string SearchMatchWholeWordsText {
            get {
                return ResourceManager.GetString("SearchMatchWholeWordsText", resourceCulture);
            }
        }
        
        /// <summary>
        ///   Looks up a localized string similar to No matches found.
        /// </summary>
        public static string SearchNoMatchesFoundText {
            get {
                return ResourceManager.GetString("SearchNoMatchesFoundText", resourceCulture);
            }
        }
        
        /// <summary>
        ///   Looks up a localized string similar to Replace all (Alt+A).
        /// </summary>
        public static string SearchReplaceAll {
            get {
                return ResourceManager.GetString("SearchReplaceAll", resourceCulture);
            }
        }
        
        /// <summary>
        ///   Looks up a localized string similar to Replace next (Alt+R).
        /// </summary>
        public static string SearchReplaceNext {
            get {
                return ResourceManager.GetString("SearchReplaceNext", resourceCulture);
            }
        }
        
        /// <summary>
        ///   Looks up a localized string similar to Toggle between find and replace modes.
        /// </summary>
        public static string SearchToggleReplace {
            get {
                return ResourceManager.GetString("SearchToggleReplace", resourceCulture);
            }
        }
        
        /// <summary>
        ///   Looks up a localized string similar to Use regular expressions.
        /// </summary>
        public static string SearchUseRegexText {
            get {
                return ResourceManager.GetString("SearchUseRegexText", resourceCulture);
            }
        }
        
        /// <summary>
        ///   Looks up a localized string similar to {0} matches.
        /// </summary>
        public static string SearchXMatches {
            get {
                return ResourceManager.GetString("SearchXMatches", resourceCulture);
            }
        }
        
        /// <summary>
        ///   Looks up a localized string similar to {0} of {1}.
        /// </summary>
        public static string SearchXOfY {
            get {
                return ResourceManager.GetString("SearchXOfY", resourceCulture);
            }
        }
    }
}
