<ResourceDictionary xmlns="https://github.com/avaloniaui"
                    xmlns:x="http://schemas.microsoft.com/winfx/2006/xaml"
                    xmlns:avaloniaedit="using:AvaloniaEdit"
                    x:ClassModifier="internal">
  <ControlTheme x:Key="{x:Type avaloniaedit:TextEditor}" TargetType="avaloniaedit:TextEditor">
    <Setter Property="Template">
      <ControlTemplate>
        <Border Background="{TemplateBinding Background}"
                BorderBrush="{TemplateBinding BorderBrush}"
                BorderThickness="{TemplateBinding BorderThickness}">
          <ScrollViewer Focusable="False"
                        Name="PART_ScrollViewer"
                        VerticalScrollBarVisibility="{TemplateBinding VerticalScrollBarVisibility}"
                        HorizontalScrollBarVisibility="{TemplateBinding HorizontalScrollBarVisibility}"
                        VerticalContentAlignment="Top"
                        HorizontalContentAlignment="Left"
                        Padding="{TemplateBinding Padding}" />
        </Border>
      </ControlTemplate>
    </Setter>
  </ControlTheme>
</ResourceDictionary>