<Application xmlns="https://github.com/avaloniaui"
             xmlns:x="http://schemas.microsoft.com/winfx/2006/xaml"
             x:Class="AvaloniaEdit.Demo.App"
             RequestedThemeVariant="Dark">
  <Application.Styles>
    <FluentTheme />
    <StyleInclude Source="avares://AvaloniaEdit/Themes/Fluent/AvaloniaEdit.xaml" />

    <!--
    <Style Selector="TextBlock.h1">
      <Setter Property="Foreground"
              Value="#212121" />
      <Setter Property="FontSize"
              Value="20" />
      <Setter Property="FontWeight"
              Value="Medium" />
    </Style>

    <Style Selector="TextBlock.h2">
      <Setter Property="Foreground"
              Value="#727272" />
      <Setter Property="FontSize"
              Value="13" />
    </Style>
    -->

  </Application.Styles>
</Application>